# -*- coding: utf-8 -*-
#
#  This file is part of Damona software
#
#  Copyright (c) 2020 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Image and Binary handlers. Provide also a Damona manager"""
import os
import sys
import pathlib
import re

import easydev

import colorlog
logger = colorlog.getLogger(__name__)


__all__ = ['Damona', 'ImageReader', 'BinaryReader', "DamonaInit"]



class DamonaInit():
    """Class to create images/bin directory from DAMONA_PATH"""
    def __init__(self):
        if "DAMONA_PATH" not in os.environ:
            logger.error("DAMONA_PATH not found in your environment. You must define "
                "it. In this shell, type 'export DAMONA_PATH=PATH_WHERE_TO_PLACE_DAMONA'")
            sys.exit(1)

        self.damona_path = pathlib.Path(os.environ["DAMONA_PATH"])
        easydev.mkdirs(self.damona_path)
        easydev.mkdirs(self.damona_path / 'envs')
        easydev.mkdirs(self.damona_path / 'images')
        easydev.mkdirs(self.damona_path / 'images' / 'damona_buffer')
        easydev.mkdirs(self.damona_path / 'bin')



class Damona():
    """Global manager to handle environments, registy, perform general
    task such as cleaning.

    """
    def __init__(self):

        if "DAMONA_PATH" not in os.environ:
            logger.error("DAMONA_PATH not found in your environment. You must define "
                         "it. In this shell, type 'export DAMONA_PATH=PATH_WHERE_TO_PLACE_DAMONA'")
            sys.exit(1)
        self.damona_path = pathlib.Path(os.environ["DAMONA_PATH"])

    def _get_config_path(self):
        return self.damona_path / "damona.cfg"
    config_path = property(_get_config_path)

    def _get_image_directory(self):
        return self.damona_path / "images"
    images_directory = property(_get_image_directory)

    def _get_environments_path(self):
        return self.damona_path / "envs"
    environments_path = property(_get_environments_path)

    def find_orphan_binaries(self):
        """Find binaries in all environments that are orphans

        By orphans, we mean that their image is not present anymore so some
        reasons.
        """
        binaries = self.get_all_binaries()
        orphans = []
        for x in binaries:
            br = BinaryReader(x)
            if br.is_image_available() is False:
                print(f"{x} image is not available. This binary does not work ")
                orphans.append(x)
        return orphans

    def get_environments(self):
        from damona.environ import Environ
        env = Environ()
        return env.environment_names

    def get_all_binaries(self):
        """Return list of all binaries in all environments"""
        from damona.environ import Environ
        env = Environ()
        binaries = [e.get_installed_binaries() for e in env.environments]
        binaries = set([x for y in binaries for x in y])
        return binaries

    def find_orphan_images(self):
        """Get images that have no binaries in any environments"""
        binaries = self.get_all_binaries()
        images = self.get_all_images()
        Nb = len(binaries)
        Ni = len(images)

        print(f"Found {Ni} images and {Nb} binaries. Checking consistencies")
        used_images = []
        for binary in binaries:
            br = BinaryReader(binary)
            used_images.append(pathlib.Path(br.image))

        used_images = set(used_images)
        Nu = len(used_images)
        No = Ni - Nu
        print(f"{Nu} images is/are used. Meaning {No} are orphans and could be removed")
        orphans = []

        for image in images:
            if image not in used_images:
                logger.info(f"{image} image not used in any environments")
                orphans.append(image)
        return orphans

    def get_all_images(self):
        """Return list of all images"""
        from damona.environ import Images
        images = Images()
        return list(images.files)


class ImageReader():
    """Manage a single image"""
    def __init__(self, name):
        """

        :param name: the input name of the image (fullpath)

        ::

            ir = ImageReader("fastqc.img")
            ir.md5
            ir.is_orphan()
            ir.name
            ir.shortname

        """
        self.filename = pathlib.Path(name)
        #.absolute())
        if self.is_valid_name() is False:
            logger.error("Invalid image name. Your input image must end in .img or .sif")
            sys.exit(1)

    def _get_short_name(self):
        return self.filename.name
    shortname = property(_get_short_name)

    def is_valid_name(self):
        pattern = r".+_(v|)\d+\.\d+\.\d+(.+|)\.(img|sif)"
        p = re.compile(pattern)
        if p.match(self.shortname):
            return True
        else:
            return False

    def _get_executable_name(self):
        pattern = r"_(v|)\d+\.\d+\.\d+(.+|)\.(img|sif)"
        p = re.compile(pattern)
        ss = p.search(self.shortname)
        guess = self.shortname[0:ss.span()[0]]
        return guess
    guessed_executable = property(_get_executable_name)

    def _get_version(self):
        pattern = r"_(v|)\d+\.\d+\.\d+(.+|)\.(img|sif)"
        p = re.compile(pattern)
        ss = p.search(self.shortname)
        version = ss.group().replace(".sif", "").replace(".img", "")
        if version[0] == "_":
            version = version[1:]
        if version[0] == 'v':
            version = version[1:]
        return version
    version = property(_get_version)

    def _get_md5sum(self):
        from easydev import md5
        md5sum = md5(self.filename)
        return md5sum
    md5 = property(_get_md5sum)

    def is_orphan(self):
        binaries = Damona().get_all_binaries()
        linked_binaries = []
        for binary in binaries:
             if self.filename == BinaryReader(binary).image:
                 linked_binaries.append(binary)
        if len(binaries) == 0:
            return True
        else:
            return False

    def is_installed(self):
        damona_path = pathlib.Path(os.environ['DAMONA_PATH'])
        if (damona_path / 'images' / self.filename.name).exists():
            return True
        else:
            return False

    def __repr__(self):
        txt = f"name: {self.filename.absolute()}\n"
        txt += f"shortname: {self.shortname}\n"
        txt += f"md5: {self.md5}\n"
        txt += f"version: {self.version}\n"
        txt += f"guessed executable: {self.guessed_executable}"
        return txt


class BinaryReader:
    """Manage a single binary"""
    def __init__(self, filename):
        """

        :param name: the input name of the binary file

        Can be use to check whether the binary is not orphan and its image is
        still available.

        """
        if isinstance(filename, str):
            filename = pathlib.Path(filename)

        with filename.open("r") as fin:

            data = [x for x in fin.readlines() if x.strip().startswith("singularity")]
            data = data[0]

            data = data.replace("${DAMONA_SINGULARITY_OPTIONS}", "") 
            try:
                image_path = data.split("exec")[1].split()[0]
            except:
                image_path = data.split("run")[1].split()[0]
                logger.warning(f"command line in {filename} uses 'run'; should be reinstalled ")

            if "DAMONA_PATH" in os.environ:
                DAMONA_PATH = os.environ['DAMONA_PATH']
                self.image = image_path.replace("${DAMONA_PATH}", DAMONA_PATH)
            else:
                self.image = image_path

    def is_image_available(self):
        if 'DAMONA_PATH' not in os.environ:
            logger.error("You must define DAMONA_PATH")
            sys.exit(1)
        damona_path = os.environ['DAMONA_PATH']
        if os.path.exists(self.image.replace("${DAMONA_PATH}", damona_path)):
            return True
        else:
            return False




# -*- coding: utf-8 -*-
#
#  This file is part of Damona software
#
#  Copyright (c) 2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/damona/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Registry and Software manager"""
import pathlib
import glob
import sys
import os
import yaml
from yaml import Loader
import packaging.version


from damona.config import Config

import colorlog
logger = colorlog.getLogger(__name__)



class Releases(dict):
    """A collection of :class:`Release`.


    """
    def __init__(self, data):
        # a collection of releases
        self._name = list(data.keys())[0]
        for version, release in data[self._name]['releases'].items():
            # enforce the keys to be strings
            # this is for special case of x.y that are read as float instead of
            # strings for e.g x.y.z
            self[str(version)] = Release(version, data)

    def _get_last_release(self):
        from packaging import version
        return max(list(self.keys()), key=lambda x: version.parse(x))
    last_release = property(_get_last_release)


class Release():
    """A Release class

    This class populates information found in the release section of a
    registry.

    A regitry looks like::

        fastqc:
            binaries: # this is the main_binaries
            x.y.z:      # a version
                download:
                md5sum:
                binaries:  # this is the extra binaries
                exclude_binaries: # exclude binaries found in the main_binaries section

    Here we have one release named x.y.z that contains information about the
    md5sum of the container, its location, and the binaries that should be
    installed.

    If binaries of the software is empty, we use the one in the specific release.
    If no binaries are set, we use the name of the software. Duplicated are
    ignored of course.

    So future release may change the list of binaries. exclude_binaries is set
    can be used to remove such binaries

    binaries can be separated by commas or spaces.

    """
    def __init__(self, version, data):
        """

        :param version: a valid x.y.z version to be found in data[name]['release']
            data is a dictionary
        params: data

        """
        self._name = list(data.keys())[0]

        try:
            self._binaries = self.split_binaries(data[self._name].get('binaries', []))
            kwargs = data[self._name]['releases'][version]
            self.download = kwargs['download']
        except Exception as err:  #pragma: no cover
            logger.error(f"Incorrect formatted registry for {self._name}" + str(err))
            sys.exit(1)

        if 'md5sum' not in kwargs:
            logger.debug(f"Missing md5sum entry in {self._name}. Please consider adding one ")
        self.md5sum = kwargs.get("md5sum", None)
        self._release_binaries = self.split_binaries(kwargs.get('binaries', []))
        self._extra_binaries = self.split_binaries(kwargs.get('extra_binaries', []))
        self._exclude_binaries = self.split_binaries(kwargs.get('exclude_binaries', []))
        self._data = data

    def _get_binaries(self):
        binaries = self._binaries + self._release_binaries + self._extra_binaries
        binaries = [x for x in binaries if x not in self._exclude_binaries]
        binaries = list(set(binaries))
        binaries = sorted(list(set(binaries)))
        if len(binaries) == 0:
            return [self._name]
        else:
            return binaries
    binaries = property(_get_binaries)

    def split_binaries(self, binaries):
        if isinstance(binaries, list):
            return binaries
        else:
            return binaries.replace(",", " ").split()

    def __repr__(self):
        binaries = ",".join(self.binaries)
        txt = f"name: {self._name}\n"
        txt += f" md5: {self.md5sum}\n"
        txt += f" binaries to be installed : {binaries}\n"
        txt += f" download from: {self.download}"
        return txt


class RemoteRegistry():
    """


    """
    def __init__(self, url):
        self.url = url
        self._read_registry()


    def _read_registry(self):

        import  urllib.request
        response = urllib.request.urlopen(self.url)
        html = response.read()
        self.rawdata = html.decode('utf-8')
        remote_registry = yaml.load(self.rawdata, Loader=Loader)
        self.data = remote_registry


class Software():
    """A class to read a given software registry

    A Software is made of :class:`Releases`. It contains also a name, a list of
    binaries either globally or per release

    """
    def __init__(self, name):
        """

        :param name: a valid name to be found in the registry. Can also be a
            dictionary with expected registry format.

        """

        if isinstance(name, dict):
            keys = list(name.keys())
            self.registry_name = keys[0]
            #: a :class:`Releases` attribute
            self.releases = self._interpret_registry(name)
        elif os.path.exists(name):
            self.registry_name = os.path.abspath(name)
            data = self._read_registry()
            #: a :class:`Releases` attribute
            self.releases = self._interpret_registry(data)
        else:
            from damona import __path__
            self.registry_name = pathlib.Path(__path__[0]) / "recipes"  / name / "registry.yaml"
            data = self._read_registry()
            self.releases = self._interpret_registry(data)

    def _read_registry(self):
        # just an alias
        regname = self.registry_name

        if os.path.exists(regname) is False: #pragma: no cover
            raise IOError(f"incorrect input filename {regname}")

        # read the yaml
        data = yaml.load(open(regname, "r").read(), Loader=Loader)
        if len(data.keys()) != 1: #pragma: no cover
            logger.error(f"{regname} must contain on single entry named after the images. ")
            sys.exit(1)

        return data

    def _interpret_registry(self, data):
        regname = self.registry_name
        releases = Releases(data)
        self._name = releases._name
        return releases

    def _get_name(self):
        return self._name
    name = property(_get_name)

    def _get_binaries(self):
        return dict([(rel, self.releases[rel].binaries) for rel in self.releases.keys()])
    binaries = property(_get_binaries)

    def _get_versions(self):
        return sorted(self.releases.keys())
    versions = property(_get_versions)

    def _get_md5(self):
        return dict([(rel, self.releases[rel].md5sum) for rel in self.releases.keys()])
    md5 = property(_get_md5)

    def check(self):
        pass

    def __repr__(self):
        txt = f"name: {self.name}\n"
        txt += f"binaries: {self.binaries}\n"
        return txt


class Registry():
    """

    The registry contains a dictionary with all images information.::

        r = Registry()
        r.registry
        r.registry['prokka_1.14.5']['download']

    """

    def __init__(self, from_url=None):

        self.config = Config().config
        if from_url:
            if from_url in self.config['urls']:
                from_url = self.config['urls'][from_url]
            else:
                assert from_url.startswith('http')
                assert from_url.endswith('registry.txt')
        self.from_url = from_url
        self.registry = {}
        self.discovery()

    def find_candidate(self, pattern):
        """Find a unique recipe within the registry.


        Note for developers: not the same as get_list"""
        candidates = [x for x in self.registry.keys() if pattern in x]
        if len(candidates) == 0: #pragma: no cover
            logger.critical(f"No image found for {pattern}. Make sure it is correct. You can use 'damona search' command")
            return None
        if len(candidates) == 1:
            return candidates[0]
        # sequana_tools_0.9.0 should return sequana_tools for the name qnd
        # 0.9.0 for the version hence the rsplit
        names = [x.rsplit(":",1)[0] for x in candidates]
        versions = [x.rsplit(":",1)[1] for x in candidates]
        version = max([packaging.version.parse(ver) for ver in versions])
        name = names[0]
        registry_name = pattern + ':' + str(version)

        return registry_name

    def discovery(self):
        """Look for recipes in the registey and populate the attributes"""
        if self.from_url:
            self._url_discovery()
        else:
            self._damona_discovery()

    def _url_discovery(self):
        self.registry = {}

        ext_reg = RemoteRegistry(self.from_url)

        for name, content in ext_reg.data.items():
            recipe = Software({name: content})
            recipe.check()
            name = recipe.name ##+ k.replace("Singularity.", "").lower()
            # we may have several releases
            for version in recipe.versions:
                name_version = recipe.name + ":" + version
                release = recipe.releases[version]
                if name_version not in self.registry:
                    if release.download is None: #pragma: no cover
                        logger.warning(f"recipe {recipe.name} has no download entry. please fill asap")
                    elif release.download.startswith("damona::"): 
                        from_url = self.config['urls']["damona"]
                        release.download = release.download.replace("damona::", from_url)
                        release.download = release.download.replace("registry.txt", "")
                    self.registry[name_version] = release
                else: #pragma: no cover
                    for kk,vv in self.registry.items():
                        print("{}: {}".format(kk, vv))
                        for kkk,vvv in self.registry[kk].items(): 
                            print(" - {}:  {}".format(kkk, vvv))
                    raise ValueError("found a duplicated name {}".format(name_version))

    def _damona_discovery(self):

        from damona.recipes import __path__
        _registry_files = glob.glob(__path__[0] + '/*/registry.yaml')

        self.registry = {}

        for registry in _registry_files:
            recipe = Software(registry)
            recipe.check()

            for version in recipe.versions:
                name_version = recipe.name + ":" + version
                release = recipe.releases[version]
                if name_version not in self.registry:
                    if release.download is None: #pragma: no cover
                        logger.warning(f"recipe {recipe.name} has no download entry. please fill asap")
                    elif release.download.startswith("damona::"):
                        from_url = self.config['urls']["damona"]
                        release.download = release.download.replace("damona::", from_url)
                        release.download = release.download.replace("registry.txt", "")
                    self.registry[name_version] = release
                else: #pragma: no cover
                    for kk,vv in self.registry.items():
                        print("{}: {}".format(kk, vv))
                        for kkk,vvv in self.registry[kk].items(): 
                            print(" - {}:  {}".format(kkk, vvv))
                    raise ValueError("found a duplicated name {}".format(name_version))

    def get_list(self, pattern=None):
        """Return list of :class:`Software` found in the registry"""
        # a name may have an underscore in it ... e.g. sequana_tools
        # in which case the singularty name is sequana_tools_0.9.0

        recipes = {}
        for name, info in self.registry.items():
            if pattern:
                if pattern.lower() in name.lower(): 
                    recipes[name] = info.download
            else:
                recipes[name] = info.download
        return recipes



import os

from damona.builders import BuilderFromDocker


def test_docker():

    import tempfile
    with tempfile.TemporaryDirectory() as td:
        bb = BuilderFromDocker()
        bb.build("docker://biocontainers/hisat2:v2.1.0-2-deb_cv1", 
            destination=td + "/test.img")



from damona import script
import damona
import subprocess
import builtins


def test_damona_app():
    from click.testing import CliRunner
    from damona.script import install, build,  env, activate, deactivate
    from damona.script import search
    runner = CliRunner()

    # isntall
    results = runner.invoke(install, ['fastqc:0.11.9'])
    assert results.exit_code == 0

    results = runner.invoke(search, ['fastqc'])
    assert results.exit_code == 0

    
    results = runner.invoke(env, [])
    # wrong
    results = runner.invoke(env, ["--create", ".dummy_test", "--delete", ".dummy_test"])
    # good
    results = runner.invoke(env, ["--create", ".dummy_test"])
    assert results.exit_code == 0
    results = runner.invoke(env, ["--create", ".dummy_test"])
    assert results.exit_code == 0
    results = runner.invoke(activate, [".dummy_test"])
    assert results.exit_code == 0
    results = runner.invoke(deactivate, [])
    assert results.exit_code == 0

    import mock
    with mock.patch.object(builtins, 'input', lambda _: 'y'):
        results = runner.invoke(env, ["--delete", ".dummy_test"])
        assert results.exit_code == 0


def test_help():
    cmd = "damona --help"
    subprocess.call(cmd.split())
    cmd = "damona pull --help"
    subprocess.call(cmd.split())


def test_activate_deactivate():
    cmd = "damona activate base"
    subprocess.call(cmd.split())
    cmd = "damona deactivate"
    subprocess.call(cmd.split())


def test_install_dryrun():
    cmd = "damona install fastqc:0.11.9 --dryrun"
    subprocess.call(cmd.split())

def test_install_wrong():
    # non existing image
    cmd = "damona install tartuffe"  
    try:
        subprocess.call(cmd.split())    
        assert False
    except:
        assert True





__all__ = ["Colors"]

class Colors:
    """

    ::

        color = Colors()
        print(color.failed("msg"))

    """
    PURPLE = "\033[95m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    def failed(self, msg):
        return self.FAIL + msg + self.ENDC

    def bold(self, msg):
        return self.BOLD + msg + self.ENDC

    def purple(self, msg):
        return self.PURPLE + msg + self.ENDC

    def underlined(self, msg):
        return self.UNDERLINE + msg + self.ENDC

    def fail(self, msg):
        return self.FAIL + msg + self.ENDC

    def error(self, msg):
        return self.FAIL + msg + self.ENDC

    def warning(self, msg):
        return self.WARNING + msg + self.ENDC

    def green(self, msg):
        return self.GREEN + msg + self.ENDC

    def blue(self, msg):
        return self.BLUE + msg + self.ENDC




from damona.registry import Software, Registry
import glob




def test_single_registry():


    from damona.recipes import __path__
    _registry_files = glob.glob(__path__[0] + '/*/registry.yaml')


    # binaries is not provided and found from the registry name
    salmon_registry = [x for x in _registry_files if "salmon" in x][0]
    sr = Software(salmon_registry)
    assert sr.releases['1.3.0'].binaries == ['salmon']

    # Sofware can take as input a full path name, a dictionary, a name
    sr = Software("salmon")
    assert sr.releases['1.3.0'].binaries == ['salmon']

    print(sr)
    sr

    # __repr__ test
    print(sr.releases['1.3.0'])
    sr.releases['1.3.0']


    # binaries is provided explicitly in the general section
    fastqc_registry = [x for x in _registry_files if "/fastqc/" in x][0]
    sr = Software(fastqc_registry)
    assert sr.releases['0.11.9'].binaries == ['fastqc']

    # binaries are provided explicitly in the releases section. Here there are
    # two releases with different binary names
    kraken_registry = [x for x in _registry_files if "kraken" in x][0]
    sr = Software(kraken_registry)
    assert 'kraken' in sr.releases['1.1'].binaries
    assert 'kraken2' in sr.releases['2.0.9'].binaries

    sr.check()
    sr.binaries
    sr.md5
    sr.releases
    sr.versions

    sr = Software("rtools")
    assert sr.releases['1.0.0'].binaries == ['R', 'Rscript']   


def test_registry():

    reg = Registry() 
    reg.get_list("fa") # only tose with fa in it
    reg.get_list() # all recipes
    reg.find_candidate("rtools")
    reg.find_candidate("fa")  # several candidate

def test_remote_registry():
    reg = Registry("https://biomics.pasteur.fr/salsa/damona/registry.txt") 
    reg.get_list()



# -*- coding: utf-8 -*-
#
#  This file is part of Damona software
#
#  Copyright (c) 2020-2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Environments manager"""
import os
import glob
import shutil
import sys
import pathlib
import math
import subprocess

from damona.common import Damona
manager = Damona()

from damona.common import BinaryReader

import colorlog
logger = colorlog.getLogger(__name__)


__all__ = ['Environ', 'Environment']


class Environment():
    """Class to handle a specific environment


    ee = environ.Environment("test1")
    ee.create_bundle("test.tar")


    """
    def __init__(self, name):
        if name == "" or name == "base":
            self.name = "base"
            self.path = manager.damona_path
        else:
            self.name = name
            self.path = manager.damona_path / f"envs/{name}/"

        if self.path.exists() is False:
            logger.error(f"Environment {self.path} does not exits")
            sys.exit(1)

    def get_installed_binaries(self):
        """Return all binaries of the environment"""
        binaries = self.path.glob("bin/*")
        binaries = [x for x in binaries if x.is_dir() is False]

        return [x for x in binaries]

    def delete(self):
        raise NotImplementedError

    def get_disk_usage(self):
        """Return virtual size of the environment if we were to
        copy/export all images """
        binaries = self.get_installed_binaries()
        S = 0
        images = self.get_images()
        for image in images:
            try:
                import os
                if 'DAMONA_PATH' not in os.environ:
                    logger.error("You must define a DAMONA_PATH")
                    sys.exit(1)
                damona_path = os.environ['DAMONA_PATH']
                image = image.replace("${DAMONA_PATH}", damona_path)
                S += os.path.getsize(image)
            except Exception as err:
                print(err)
                logger.error(f"Could not check {image}")
        return S

    def get_images(self):
        binaries = self.get_installed_binaries()
        images = []
        for binary in binaries:
            br = BinaryReader(binary)
            if br.image not in images:
                images.append(br.image)
        return images

    def __repr__(self):
        N = len(self.get_installed_binaries())
        M = len(self.get_images())

        txt = f"Contains {N} binaries from {M} images."
        disk = self.get_disk_usage()
        import math
        env_size = math.ceil(disk / 1e6)

        txt += f" Disk usage is {env_size}Mb"
        return txt

    def rename(self, newname):
        raise NotImplementedError
        if newname in []:
            pass
        else:
            logger.info(f"Renaming {self.path} into {newname}")
            ff = pathlib.Path(self.path)
            ff.rename(newname)

    def create_bundle(self, output_name=None, exclude=None):
        if output_name is None:
            output_name = self.name


        if exclude:
            exclude = list(self.path.glob(f"bin/{exclude}"))
        else:
            exclude = []

        # all binaries
        binaries = self.get_installed_binaries()
        binaries = [x.absolute() for x in binaries if x not in exclude]
        binaries = sorted(binaries)

        # all containers
        images = [pathlib.Path(x) for x in self.get_images()]

        import tarfile
        archive = tarfile.open(f"damona_{output_name}.tar", "w")
        for filename in binaries:
            logger.info(f"Adding {filename}")
            archive.add(filename, arcname=f"bin/{filename.name}")
        for filename in images:
            logger.info(f"Adding {filename}")
            archive.add(filename, arcname=f"images/{filename.name}")
        archive.close()

        logger.info(f"Saved environment {self.name} into damona_{output_name}.tar")


class Images():
    def __init__(self):
        self.images_dir = manager.images_directory

    def __len__(self):
        return len(list(self.files))

    def _get_images(self):
        return self.images_dir.glob("*.img")
    files = property(_get_images)

    def get_disk_usage(self, frmt="Mb"):
        env_size = sum(os.path.getsize(f) for f in self.files if os.path.isfile(f))
        if frmt == 'Mb':
            return math.ceil(env_size/1e6)
        else:
            return env_size


class Environ():
    """Class to deal with the damona environments"""

    def __init__(self):
        self.images = Images()

    @staticmethod
    def get_current_env():
        if 'DAMONA_ENV' not in os.environ:
            path = manager.damona_path
        else:
            path = pathlib.Path(os.environ['DAMONA_ENV'])
        return path

    def _get_N(self):
        return len(self.environments)
    N = property(_get_N)

    def _get_envs(self):
        path_envs = manager.damona_path / "envs"
        envs = [Environment("base")] + [Environment(x.name) for x in path_envs.iterdir()]
        return envs
    environments = property(_get_envs)

    def _get_env_names(self):
        envs = os.listdir(manager.environments_path)
        envs = ['base'] + [Environment(x).name for x in envs ]
        return envs
    environment_names = property(_get_env_names)

    def delete(self, env_name):
        env_path = manager.environments_path / env_name
        if os.path.exists(env_path) is False:
            logger.error("{} does not exists".format(env_path))
        else:
            try:
                os.rmdir(env_path)
            except OSError:
                logger.warning("Will delete all contents of {}. Although this concerns only aliases you will lose your environement".format(env_path))
                ret = input("Are you sure you want to proceed ? (N/y)")
                if ret == 'y':
                    shutil.rmtree(env_path)


    def _env_in_path(self, env_name):
        PATH = os.environ['PATH']
        paths = PATH.split(":")
        for x in paths:
            if str(manager.damona_path / "envs" / env_name / "bin") == x:
                return True
        return False

    def activate(self, env_name=None):
        # Do not change the print statement here below. They are used by
        # damona.sh
        if env_name not in self.environment_names + ['base']:
            logger.error(f"invalid environment:  {env_name}. Please use 'damona env' to get the list")
            sys.exit(1)

        if self._env_in_path(env_name) is True:
            logger.warning(f"damona environment {env_name} is already in your PATH. nothing done")
            return 

        if env_name == "base":
            env_path = manager.environments_path.parent
            print('    export DAMONA_ENV={};'.format(env_path))
            print('export PATH={}/bin:${{PATH}}'.format(env_path))
        else:
            env_path = manager.environments_path / env_name
            print('    export DAMONA_ENV={};'.format(env_path))
            print('export PATH={}/bin:${{PATH}}'.format(env_path))
        logger.info(f"# Added damona path ({env_path}) in your PATH")

    def deactivate(self, env_name=None):
        # we deactivate the latest activated damona environment only.
        # can be called several times. If called too many times,
        # we set the main damona environment (base) as default
        PATH = os.environ['PATH']
        paths = PATH.split(":")

        found = False   # this one is the one to deactivate (to ignore)
        newPATH = []
        for path in paths:
            #logger.info(f"# {env_name} {path}")

            # if an env_name is provided, it may be removed several times. 
            # if not provided, 
            if env_name and str(manager.damona_path / "envs" / env_name / "bin") == path:
                logger.info(f"# Found damona path ({path}), now removed from your PATH")
                found = True
            elif env_name is None and f"/damona/" in path and found is False:
                logger.info(f"# Found damona path ({path}), now removed from your PATH")
                found = True
            else: # keep track of the DAMONA_ENV.
                newPATH.append(path)

        if found is False:
            logger.info("# no more active damona environment in your path. Use 'damona activate ENVNAME'")
        #
        newPATH = ":".join(newPATH)
        print('export PATH={}'.format(newPATH))

    def create(self, env_name, force=False):
        if env_name == "base":
            logger.critical("base is a reserved name for environment. Cannot be created")
            sys.exit(1)

        env_directory = manager.damona_path
        env_path = env_directory / "envs" / env_name
        if os.path.exists(env_path) and force is False:
            logger.critical("{} exists already".format(env_path))
            sys.exit(0)
        else:
            try:
                os.mkdir(env_path)
                os.mkdir(env_path / "bin")
                logger.info("Created {} in {}".format(env_name, env_directory))
            except: #pragma: no cover
                pass # if already created, error are caught here

    def create_from_bundle(self, env_name, bundle, force=False):
        if env_name in self.environment_names:
            logger.warning(f"{env_name} exists already.")
            if force is False:
                logger.critical(f"To recreate, you must delete it using 'damona env --delete {env_name}'")
                sys.exit(0)
            else:
                logger.warning(f"You used --force, so overwritting existing environment")

        # if it does not exsits or force is True
        self.create(env_name, force=force)

        env_directory = pathlib.Path(manager.damona_path / "envs" / env_name)

        import tarfile
        archive = tarfile.open(bundle)
        for x in archive.getmembers():
            if x.name.startswith("bin"): # this is a binary
                archive.extract(x, path=env_directory)
            elif x.name.startswith("images"): # a contaier
                archive.extract(x, path=env_directory)
                from damona.common import ImageReader
                temp_image = ImageReader(env_directory / x.name)
                if temp_image.is_installed():
                    target_image = ImageReader(manager.damona_path / x.name)
                    if target_image.md5 == temp_image.md5:
                        logger.info(f"- Image {x.name} exists already with same md5sum. Skipped copy.")
                    else:
                        logger.warning(f"- Image {x.name} exists already with diffrent "
                            " md5sum. This should not happen with damona images."
                            " You will need to copy manually if required")
                else:
                    logger.info(f"- Copying {x.name} into {manager.damona_path}/images")
                    temp_image.filename.rename(manager.damona_path / "images"/ temp_image.filename.name)
            else: # we should not enter here
                raise NotImplementedError

        # finally, we remove all images since they are redundant or have
        # been copied
        def rm_tree(pth):
            pth = pathlib.Path(pth)
            for child in pth.glob('*.img'):
                if child.is_file():
                    child.unlink()
                else:
                    rm_tree(child)
            pth.rmdir()
        to_delete = (env_directory / "images").absolute()
        logger.info(f"Removing temporary {to_delete} directory")
        rm_tree(env_directory / "images")

    def copy(self):
        """Copy an environment"""
        raise NotImplementedError



from damona.colors import Colors



def test_colors():
    c = Colors()    
    c.failed("test")
    c.bold("test")
    c.purple("test")
    c.underlined("test")
    c.fail("test")
    c.error("test")
    c.warning("test")
    c.green("test")
    c.blue("test")



# -*- coding: utf-8 -*-

###########################################################################
# Damona is a project to manage reproducible containers                  #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2020-2021  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# Damona is free software: you can redistribute it and/or modify          #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# Damona  is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
""".. rubric:: Standalone application dedicated to conversion"""
import click
import glob
import sys
import pathlib
import os
import subprocess

from damona  import version
from damona import Damona
manager = Damona()


__all__ = ["main"]

from damona import logger
logger.level = 10



CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS)
@click.option("--level", default="INFO",
    help="""Set level information to DEBUG, INFO, WARNING, CRITICAL, ERROR. Use e.g., "damona --level INFO command" """)
@click.version_option(version=version)
def main(level):
    """Damona is a singularity container manager to ease the deployment of
    reproducible working environments. It is to singularity container what conda
    is to packaging.

    You can create and activate a Damona environment using

    \b
        damona env --create env1
        damona activate env1

    Once an environment is activated, you can install a Damona-registered image
    (otherwise it is installed in the default environement):

        damona install fastqc

    or from an online registry (given a URL)::

        damona install fastqc:0.11.9 --from-url https://.../registry.txt

    You may build and install a local image (not recommended; for developers):

    \b
        damona build Singularity.test_1.0.0
        damona install test_1.0.0.img --binary-name fastqc

    Images can be built from docker:

        damona build docker://whatever

    An environment can be exported and installed on another computer as follows

        damona export test1 --name myenv
        # on the remote computer:
        damona import myenv

    \b
    More information on https://damona.readthedocs.io.
    Please report issues on https://github.com/cokelaer/damona
    Contact: Thomas Cokelaer at pasteur dot fr
    """
    from damona.colors import Colors
    ##### !!!!!!!!!!!! this function cannot print anything because the damona
    # activate command prints bash commands read by damona.sh
    #print(Colors().blue("# Welcome to Damona (Container Manager)"))
    from damona import logger
    logger.setLevel(level)


@main.command()
@click.argument('filename', required=True, type=click.STRING)
@click.option('--destination', default=None,
    help="Not implemented yet")
@click.option('--force', is_flag=True,
    help="add --force option")
def build(**kwargs): #pragma: no cover
    """Build a container from dockerhub, singularity file or damona recipes.

    Note that to use this command, you must have sudo permissions.
    If not, you will need to download existing images. See the "damona install
    command".

    You can build a singularity image from a local singularity file. Note that
    your Singulary recipes must follow the naming convention
    Singularity.NAME_x,y,z

    \b
        # a local recipes (recipes must have a version)
        damona build Singularity.salmon_1.3.0

    You may build an image from a singularity recipes to be found in Damona
    itself. In such case, the name and version are enough. Siuch recipes can be
    listed using "damona list"

        damona build salmon:1.3.0

    You may also build image from a docker image to be found on docker hub:

    \b
        damona build docker://biocontainers/bowtie2:v2.4.1_cv1
        damona build docker://kapeel/hisat2

    TO be implemented: from singularity hub or sylabs.io

    """
    logger.debug(kwargs)
    filename = kwargs['filename']
    force = kwargs['force']
    destination = kwargs['destination']

    if destination:
        if destination.endswith(".img") is False:
            logger.warning("You should end your image with the .img extension")
        if "_" not in destination:
            logger.warning("You should name your image as NAME_x.y.z")

    if os.path.exists(filename) and os.path.isdir(filename) is False:
        # TODO check that it starts with Singularity
        # local recipes ?
        from damona.builders import BuilderFromSingularityRecipe
        builder = BuilderFromSingularityRecipe()
        builder.build(filename, destination=destination, force=force)
    elif kwargs['filename'].startswith("docker://"):
        from damona.builders import BuilderFromDocker
        builder = BuilderFromDocker()
        filename = filename.replace("docker://","")
        builder.build(filename, destination=destination, force=force)
    else: # could be a damona recipes
        logger.info("Not a docker URL, nor a local file.")


@main.command()
@click.option('--pattern', default=None,
    help="restrict the output list keeping those with this pattern")
@click.option('--from-url', help="""download registry from remote URL. The URL must
contain a registry.txt as explained on damona.readthedocs.io. If you just type
'damona', the URL will be biomics.pasteur.fr/drylab/damona/registry.txt""")
@click.option('--url', help="""alias to --from-url""")
def available_images(**kwargs):
    """List all available images from Damona.

    # TODO: list in conda is the list of installed paclages. we should rename this comamnd
    # into e.g.,

     all available images:

        damona list

    List all available images available at a given URL:

        damona list --from-url https://biomics.pasteur.fr/drylab/damona/registry.txt

    Same command as above (alias) using the short name::

        damona list --from-url damona

    """
    from damona.registry import Registry
    if kwargs['url']:
        url = kwargs['url']
    elif kwargs['from_url']:
        url = kwargs['from_url']
    else:
        url = None
    modules = Registry(from_url=url).get_list(pattern=kwargs['pattern'])
    names = sorted(list(modules.keys()))
    downloads = [modules[x] for x in names]
    print("name                 Download location")
    for k,v in zip(names, downloads):
        print(f"{k:20} [{v}]")


#TODO --rename

@main.command()
@click.option('--create', type=click.STRING,
    help="""create a new environment""")
@click.option('--delete', type=click.STRING,
    help="""Delete an existing environment""")
@click.option('--disk-usage', is_flag=True,
    help="Prints disk usage of each Damona environments")
@click.option('--from-bundle', type=click.STRING,
    help="a bundle file create with 'damona export' command")
@click.option('--force', is_flag=True,
    help="""When creating an environment with --from-bundle, rewrite binaries and
images even though the environment exists.""")
def env(**kwargs):
    """Get information about Damona environments.

    Print the current environments and available environments:

        damona env

    Check the disk usage of each environments::

        damona env --disk-usage

    You can also create an environment and install a saved on in it:
    
    \b
        damona export test1
        damona env --create copy_test1 --from-bundle damona_test1.tar

    """
    logger.debug(kwargs)
    from damona import Environ
    envs = Environ()

    if kwargs['disk_usage']:
        import math
        from damona import Environ
        envs = Environ()
        #base = envs.get_current_env() + "/images/*img"
        #env_size = sum(os.path.getsize(f) for f in glob.glob(base) if os.path.isfile(f))
        N = len(envs.images)
        usage = envs.images.get_disk_usage()
        #env_size = math.ceil(env_size/1e6)
        print(f"Found {N} images: Images usage: {usage}.Mb")
    elif kwargs['create'] is None and kwargs['delete'] is None:
        print(f"There is currently one base (default) environment and {envs.N-1} user environments.")
        if envs.N !=0:
            for this in envs.environments:
                name = this.name
                print(f"{name} -  {this}")
        current_env = envs.get_current_env()
        logger.info("""Your current env is {}.""".format(current_env))
    elif kwargs['create'] and kwargs['delete']: # mutually exclusive
        logger.error("you cannot use --delete and --create together")
    elif kwargs['delete']:
        envs.delete(kwargs['delete'])
    elif kwargs['create']:
        if kwargs['from_bundle']:
            envs.create_from_bundle(kwargs['create'],
                bundle=kwargs['from_bundle'], force=kwargs['force'])
        else:
            envs.create(kwargs["create"])


@main.command()
@click.argument('name', required=True, type=click.STRING)
# TODO:
# @click.argument('--all', e, type=click.STRING)
def activate(**kwargs):
    """Activate a damona environment.


    The main Damona environment can be activated using

        damona activate base

    Then, activation of a specific environment is done as :

        damona activate my_favorite_env

    """
    # logger.debug(kwargs)
    from damona import Environ
    env = Environ()
    env.activate(kwargs['name'])


@main.command()
@click.argument('name', required=False, type=click.STRING, default=None)
def deactivate(**kwargs):
    """Deactivate the current Damona environment.

    deactivate the current environement:

        damona deactivate

    """
    logger.debug(kwargs)
    from damona import Environ
    env = Environ()
    env.deactivate(kwargs['name'])


@main.command()
@click.argument('image', required=True, type=click.STRING)
@click.option('--force-image', is_flag=True, help="Replaces existing image.")
@click.option('--force', is_flag=True, help="Replaces images and binaries.")
@click.option('--force-binaries', is_flag=True, help="Replace binaries.")
@click.option('--from-url', help="""download image from a remote URL. The URL must
contain a registry.txt as explained on https://damona.readthedocs.io""")
@click.option('--url', help="""alias to --from-url""")


@click.option('--binaries', default=None,
    help="""If not provided, we assume this is an executable singulatrity and its name is the binary name
    """)
def install(**kwargs):
    """Download and install an image and its binaries.

    The main idea of Damona is to download and install images already built for
    you and to install binaries to be found inside the image. An image has a
    name and a version and can be installed as:

        damona install NAME:version

    If the version is omitted, the latest version is installed. Therefore, the
    following two commands are equivalent (as of march 2021):

    \b
        damona install fastqc
        damona install fastqc:0.11.9

    You may also install a local image. In such case, you must name your
    image with a version (e.g. fastqc_0.11.9.img). We assume that name of
    the image is the binary name, however, binaries can be set manually:

    \b
        damona install fastqc_0.4.2.img
        damona install test_0.4.2.img --binary fastqc

    By convention, images must follow the pattern NAME_[v]x.y.z[_info].img
    where the [v] character is optional and the [_info] is an optional _
    character followed by any text. Finally, the extension can be .img or .sif

    Images are installed in the ~/.config/damona/images directory. It is the
    DAMONA_PATH environment variable. Therefore, you can redefine this variable
    to install images elsewhere.

    Finally, you may have images online on a website. To install such images, use
    the --from-url/--url (sse developer guide for details).
    """
    logger.debug(kwargs)

    # url
    if kwargs['url']:
        url = kwargs['url']
    elif kwargs['from_url']:
        url = kwargs['from_url']
    else:
        url = None

    image_path = pathlib.Path(kwargs['image']).absolute()

    force = kwargs['force']
    force_image = kwargs['force_image']
    force_binaries = kwargs['force_binaries']
    if force:
        force_image, force_binaries = True, True

    if kwargs['binaries']:
        if "," in kwargs['binaries']:
            binaries = kwargs['binaries'].split(",")
        else:
            binaries = [kwargs['binaries']]
    else:
        binaries = None

    if os.path.exists(image_path) is False:
        if kwargs['from_url']:
            url = kwargs['from_url']
            logger.info(f"Installing from online registry  (url: {url})")
        else:
            logger.info("Installing from Damona registry")
        from damona.install import RemoteImageInstaller
        p = RemoteImageInstaller(kwargs['image'],
                from_url=kwargs['from_url'], cmd=sys.argv, binaries=binaries)
        if p.is_valid():
            p.pull_image(force=force_image)
            p.install_binaries(force=force_binaries)
        else:
            logger.critical("Something wrong with your image/binaries. See message above")
    else:
        # This install the image and associated binary/binaries
        from damona import environ
        env = environ.Environ()
        cenv = env.get_current_env()
        logger.info(f"Installing local container in {cenv}")
        from damona.install import LocalImageInstaller
        lii = LocalImageInstaller(image_path, cmd=sys.argv, binaries=binaries)
        if lii.is_valid():
            lii.install_image(force=force_image)
            lii.install_binaries(force=force_binaries)
        else:
            logger.critical("Something wrong with your image/binaries. See message above")


'''
@main.command()
@click.argument('name', required=True, type=click.STRING)
@click.option("--from-image", required=True)
@click.option("--env", required=True)
def add_binary(**kwargs):
    """Create a binary linked to an existing image in a specfici environment

        conda add-binary fastqc --from-image test_0.4.0.img --env ENVNAME
    """
    logger.debug(kwargs)
    raise NotImplementedError
    #TODO check that env exists
    env = kwargs['env']
    if env not in manager.environment_names:
        logger.error(f"Environment {env} not found. Check valid names using 'damona env'")
        sys.exit()
    raise NotImplementedError
'''

@main.command()
@click.option("--remove", is_flag=True,
              help="--remove the binary and image orphans in all environments ")
def clean(**kwargs):
    """Remove orphan images and binaries from all environments.

    This commnd finds images that have no more binary in any environment.
    This may happen with prior version of a given binary.

    Note that if you want to remove an environmnt, just delete it or use:

    conda env --delete NAME
    """
    logger.debug(kwargs)
    from damona import Damona
    dmn = Damona()

    # First we deal with orphans from the binaries directories
    orphans = dmn.find_orphan_binaries()
    if len(orphans) == 0:
        logger.info("No orphans binary found")

    if kwargs['remove']:
        for x in orphans:
            os.remove(os.path.expanduser(x))
            logger.info(f"Removed {x}")
    else:
        logger.warning("Please use --remove to confirm that you want to remove the orphans")

    # Second, we find images that have no more binaries
    orphans = dmn.find_orphan_images()
    if len(orphans) == 0:
        logger.info("No orphan images found")
    else:
        logger.info(f"Found {len(orphans)} orphans.")

    if kwargs['remove']:
        for x in orphans:
            answer = input(f"You are going to delete this image: {x}. Are you sure ? (yes/no)")
            if answer == "yes":
                os.remove(os.path.expanduser(x))
                logger.info(f"Removed {x}")
            else:
                logger.info(f"skipped deletion of {x}")
    else:
        logger.warning("Please use --remove to confirm that you want to remove the orphans")

@main.command()
@click.argument('pattern', required=True, type=click.STRING)
@click.option('--from-url', help="""Set the online registry file to search for a
given container. See damona.readthedocs.io for information on how to write this
file . Example is available on https://biomics.pasteur.fr/salsa/damona/registry.txt""")
@click.option('--url', help="""alias to --from-url""")
def search(**kwargs):
    """Search for a container given a pattern in damona registry.

    By default searches for a software in Damona only:

        damona search fastqc

    IF you want to list all software and their versions, just type:

        damona search "*"

    One can also search in an online registry:

        damona search fastqc --url https://biomics.pasteur.fr/salsa/damona/registry.txt

    You may define aliases to URLs in your ~/.config/damona/damona.cfg file to
    make it easier:

        damona search fastqc --url damona

    """
    logger.debug(kwargs)
    from damona.registry import Registry
    if kwargs['url']:
        url = kwargs['url']
    elif kwargs['from_url']:
        url = kwargs['from_url']
    else:
        url = None

    if kwargs['pattern'] == "*":
        pattern = None
    else:
        pattern = kwargs['pattern']
    modules = Registry(from_url=url).get_list(pattern=pattern)

    for mod in modules:
        print(mod)



@main.command()
@click.argument('environment', required=True, type=click.STRING)
def info(**kwargs):
    """Print information about a given environement.

    The default environment is called 'base'.

    \b
        damona info base
        damona info test1

    Images abd binaries available are shown
    """
    logger.debug(kwargs)
    envname = kwargs['environment']
    from damona.environ import Environ
    manager = Environ()

    x = [ee for ee in manager.environments if ee.name == envname]
    if len(x) == 0:
        logger.error(f"Environment '{envname}' does not exist. Use 'damona env' to get the list")
        sys.exit(1)
    else:
        environ = x[0]
        print(environ)
        print("Images:")
        for item in sorted(environ.get_images()):
            print(" - " + pathlib.Path(item).name)
        print("Binaries:")
        for item in sorted(environ.get_installed_binaries()):
            print(" - " + pathlib.Path(item).name)

@main.command()
@click.argument('environment', required=True, type=click.STRING)
@click.option("--exclude", default=None,
              help="--exclude 'bowtie*' ")
def export(**kwargs):
    """Create a bundle of a given environment.

    the following command copies all binaries from an environement and their
    associated images into a tar ball file named after the
    environment.

        damona export test1 --exclude "bowtie*"

    This create a bundle named damona_test1.tar. You can then create a new 
    environment starting from this bundle:

        damona env --create TEST1 --from-bundle damona_test1.tar

    """
    logger.debug(kwargs)
    environment = kwargs['environment']
    exclude = kwargs['exclude']

    # TODO This should be based on the binaries of the environment, not the images
    # to do so, we'll need an installed.txt file

    from damona import Environment
    envname = kwargs['environment']
    env = Environment(envname)
    env.create_bundle(exclude=exclude)
    logger.info(f"Saved environment into {environment}.tar")
    logger.info(f"Use this command to create a new environment: damona env --create test --from-bundle {environment}.tar")






if __name__ == "__main__": #pragma: no cover
    main()




# -*- coding: utf-8 -*-
#
#  This file is part of Damona software
#
#  Copyright (c) 2020-2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#          <d.desvillechabrol@gmail.com>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""The Damona configuration"""
import os
import sys
import pathlib

import colorlog
logger = colorlog.getLogger(__name__)



# list of URLs where to find registry and their aliases
urls = {
    'damona': "https://biomics.pasteur.fr/salsa/damona/registry.txt"
}


class Config():
    """A place holder to store our configutation file

    ::

        [urls]
        damona=https://....
        url1=https://....

    """
    def __init__(self):
        from easydev import CustomConfig
        configuration = CustomConfig("damona", verbose=True)

        #  let us add a damona.cfg in it. This will store URLs to look for singularities
        # This is done only once to not overwrite user options
        self.user_config_dir = pathlib.Path(configuration.user_config_dir)

        self.config_file = self.user_config_dir / "damona.cfg"
        if self.config_file.exists() is False: #pragma: no cover
            with open(self.config_file, "w") as fout:
                fout.write("[general]\n")
                fout.write("show_init_warning_message=False\n\n")
                fout.write("[urls]\n")
                for k,v in urls.items():
                    fout.write("{}={}".format(k,v))
        else:
            logger.debug("damona.cfg file exists already. Reading it")

        # read the config
        self.read()

        # create the shell script once for all
        created = self.add_shell()
        if created:
            logger.critical("Please start a new shell to benefit from "
                "the configuration file and activate/deactivate command")
            #sys.exit(1)

    def read(self):
        from configparser import ConfigParser
        config = ConfigParser()
        config.read_file(open(self.config_file))
        self.config = config

    def add_shell(self): #pragma: no cover  ; this is executed only if config does not exists
        #  let us add a damona.cfg in it. This will store URLs to look for singularities
        # This is done only once to not overwrite user options
        if os.path.exists(self.user_config_dir / "damona.sh") is False:
            logger.info("adding a damona.sh in your DAMONA_PATH")
            _damona_config_path = self.user_config_dir
            logger.warning(f"Creating damona.sh file in {_damona_config_path}. ")
            import damona.shell
            shell_path = damona.shell.__path__[0]
            with open(shell_path + os.sep + "damona.sh", "r") as fin:
                with open(_damona_config_path / "damona.sh", "w") as fout:
                    fout.write(fin.read())
            return True
        else:
            return False



from damona.common import *

import os
import pytest

def test_no_path(monkeypatch):

    monkeypatch.delenv("DAMONA_PATH", raising=False)
    try:
        DamonaInit()
        assert False
    except:
        assert True
    try:
        Damona()
        assert False
    except:
        assert True

def test_path():
    #os.environ['DAMONA_PATH'] = '/tmp'
    d = Damona()
    d.config_path

    d.find_orphan_binaries()
    d.get_environments()
    d.find_orphan_images()




import pkg_resources
import os
import pathlib
import colorlog

try:
    version = pkg_resources.require("damona")[0].version
except Exception:   #pragma: no cover
    version = ">=0.8.3"


# The logger mechanism is here:
handler = colorlog.StreamHandler()
handler.setFormatter(colorlog.ColoredFormatter(
    '%(log_color)s%(levelname)s: [%(name)s,l %(lineno)s]: %(message)s'))
logger = colorlog.getLogger("damona")
logger.addHandler(handler)


# Here we create a persistent config directory in the Home of the user.
# This is a small file.
from damona.config import Config
Config()



 # Some information for the users
if "DAMONA_EXE" not in os.environ: #pragma: no cover
    #logger.critical("No DAMONA_EXE environment variable found")
    logger.critical("Damona binaries are installed in ~/.config/damona/bin by default")
    logger.critical("You may install them in specific environments and activate/deactivate"
        " the environments to you convenience.")
    logger.critical("You will need to set the PATH manually so that you may "
        "find binaries in ~/.config/damona/bin or one of the environment "
        "in ~/.config/damona/envs")
    logger.critical("To remove this message, and benefit from the "
        "activate/deactivate command, add this line in your .bashrc\n"
        "source ~/.config/damona/damona.sh\n")


# Based on the previous config path, we may add images, environments and
# binaries if  DAMONA_PATH is not defined. If DAMONA_PATH is redefined,
# the following call creates the images/ envs/ bin/ directories
from damona.common import DamonaInit
DamonaInit()


# The user/developer API
from damona.registry import Registry
from damona.environ import Environ, Environment
from damona.common import Damona



from damona.environ import Environment
from damona.environ import Environ
import damona
import builtins



def test_no_var(monkeypatch):
    monkeypatch.delenv("DAMONA_ENV", raising=False)
    try:
        Environ()
        assert False
    except:
        assert True



def test_environ():
    env = Environ()
    env.N
    env.environments
    env.create(".dummy_test")
    import mock
    with mock.patch.object(builtins, 'input', lambda _: 'y'):
        env.delete(".dummy_test")


    env.get_current_env()
    env.environment_names
    env.activate("base")
    env.deactivate()


def test_environment_no_var(monkeypatch):
    monkeypatch.delenv("DAMONA_PATH", raising=False)
    e = Environment("base")
    try:
        e.get_disk_usage()
        assert False
    except:
        assert True

def test_environment():
    e = Environment("base")
    e.get_installed_binaries()
    #from tempfile import TemporaryFile
    #with TemporaryFile() as fout:
    #    e.create_bundle(output_name=fout.name)

    try:
        e = Environment("base_does_not_exist")
        assert False
    except:
        assert True





# -*- coding: utf-8 -*-
#
#  This file is part of Damona software
#
#  Copyright (c) 2020 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Tools to install images/binaries"""
import os
import sys
import pathlib
import packaging.version
import shutil
import time
import subprocess



from spython.main import Client
from damona import Registry
from damona import Environ
from damona.common import ImageReader

import os
DAMONA_PATH = os.environ['DAMONA_PATH']



import colorlog
logger = colorlog.getLogger(__name__)


__all__ = ['LocalImageInstaller', 'RemoteImageInstaller']


class CMD():
    def __init__(self, cmd):
        self.cmd = cmd
        from damona import version
        self.version = version

    def __repr__(self):
        assert self.cmd[0].endswith("damona")
        cmd = "damona " + " ".join(self.cmd[1:])
        return cmd

class ImageInstaller():

    def _are_binaries_findable(self):
        # TODO add sanity check that stops the installation if a failure occurs
        for binary in self.binaries:
            # is it found in the image ? without -v may return error code
            # so we first test with --version option but some tools will use -v
            # so we test --version and -v cases. Finally, if there is no
            # version, let us just try without arguments. Most probably an error
            # code is returned. ut if not found at all, a 'executable file not
            # found' message should be returned by bash. Not ideal but will do
            # for now.
            cmd1 = f"singularity exec {self.input_image.filename} {binary}"
            cmd2 = f"singularity exec {self.input_image.filename} {binary} -v"
            cmd3 = f"singularity exec {self.input_image.filename} {binary} --version"
            # ideally we should use commad but not present on all systems....
            # cmd = f"singularity exec {self.input_image.filename} command -v {binary}"

            status = subprocess.Popen(cmd3, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            status.wait()
            if status.returncode == 0:
                logger.info(f"'{binary}' binary found in the container. Planned to be installed")
                continue

            status = subprocess.Popen(cmd2, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            status.wait()
            if status.returncode == 0:
                logger.info(f"'{binary}' binary found in the container. Planned to be installed")
                continue

            status = subprocess.Popen(cmd1, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            status.wait()
            if status.returncode == 0:
                logger.info(f"'{binary}' binary found in the container. Planned to be installed")
                continue
            else:
                if status.stderr is None:
                    logger.critical("Unknown error due to errocode != 0 but not error message reported")
                    #FIXME. not sure this ever happens
                    return False

                error = status.stderr.read()
                if 'executable file not found' in error.decode():
                    logger.critical(f"{binary} executable not available in the image {self.input_image.filename}")
                    return False
                else:
                    # Other type of error could simply be help printed to stderr...
                    # let us assume this is the case but let us print a warning
                    # for future debugging
                    logger.warning(f"Error while testing presence of {binary}. Most probably related to missing arguments. Here is the error code returned: \n\n{error}\n\n Proceed")
        return True



class LocalImageInstaller(ImageInstaller):

    def __init__(self, image_name, cmd=None, binaries=None):
        # must be a valid name with version, name, etc so that executable name
        # can be guessed
        super(LocalImageInstaller, self).__init__()

        self.input_image = ImageReader(image_name)
        self.target = pathlib.Path(DAMONA_PATH) / "images" / self.input_image.shortname
        self.images_directory = pathlib.Path(DAMONA_PATH) / "images"
        self.cmd = cmd

        # If no binaries are provided, we will guess from the name of the image
        # the unique possible binary to be found in the image.
        if binaries is None:
            self.binaries = [self.input_image.guessed_executable]
        else:
            self.binaries = binaries

        self.image_installed = False

    def is_valid(self):

        # let us check the  presence of the binaries in the image
        status = self._are_binaries_findable()
        return status

    def install_image(self, force=False):
        if (self.images_directory / self.input_image.shortname).exists():
            from easydev import md5
            md5_target = md5(self.images_directory / self.input_image.shortname)
            if md5_target == self.input_image.md5:
                logger.info("Image with same md5 exists already. No need to copy")
                # image is not copied since it is already installed
                self.image_installed = True
            else:
                if force:
                    self.copy()
                else:
                    logger.warning(f"{self.input_image.filename} exists with different md5sum. Please use --force to overwrite")
        else:
            self.copy()

    def copy(self):

        logger.info(f"Copying {self.input_image.filename} into {self.images_directory}")
        shutil.copy(self.input_image.shortname, self.images_directory)
        with open(self.images_directory / "history.log", "a+") as fout:
            if self.cmd:
                cmd = CMD(self.cmd).__repr__()
                fout.write(f"{time.asctime()}: {cmd}\n")
        # if the image has been copied properly, we set this flag to True so
        # that binaries can be installed
        self.image_installed = True

    def install_binaries(self, force=False):

        if self.image_installed:
            bininst = BinaryInstaller(self.binaries, self.input_image.filename)
            bininst.install_binaries(force=force)
        else:
            logger.critical("The container image has not been installed. So, binaries cannot be installed either")



class RemoteImageInstaller(ImageInstaller):
    """Manager to download container images

    A valid singularity image must have the following name::

        Singularity.NAME_x.y.z
        Singularity.NAME_SUFFIX_x.y.z

    Meaning underscore can be part of the name.

    User interface is simplified into::

        NAME:x.y.z
        NAME_SUFFIX:x.y.z

    The following command will download the fastqc singularity version 0.11.9
    from the damona website.::

        damona install fastqc:0.11.9

    If you have a damona registry on your website, you can download from a URL:

        damona install fastqc:0.11.9 --from-url https://yourwebsite

    """
    def __init__(self, image_name, binaries=None, from_url=None, cmd=None):
        super(RemoteImageInstaller, self).__init__()


        self.image_name = image_name
        self.registry = Registry(from_url=from_url)
        self.images_directory = pathlib.Path(DAMONA_PATH) / "images"

        self.from_url = from_url
        self.cmd = cmd
        self.binaries = binaries
        self.image_installed = False

    def is_valid(self):

        return True

    def pull_image(self, output_name=None, force=False):

        self.image_installed = False

        # the name must be valid. we use _ to separate name and version for the
        # singularity filemane but the image have a tag (version) and name
        # separated by a :

        # Our registry uses underscore. The output filename should use _ as well
        # However, for users and image stored on sylabs or singularity hub, we
        # use the character :

        # let us first retrieve the image and save it in a temporary file. 
        # we keep the original name by only replacing the : with underscore so
        # that we can use ImageReader class to help us later on.
        if ":" in self.image_name:
            logger.info(f"Looking for {self.image_name}...")
            # e.g. fastqc:0.11.9
            registry_name = self.image_name #.replace(":", "_")

            if self.image_name not in self.registry.registry.keys():
                logger.critical("invalid image name provided: {}. type 'damona list'".format(self.image_name))
                guess = self.image_name.split(":")[0][0:4]
                # here we reverse x to start from the end and replace the last _
                # by :
                guesses = [x[::-1].replace("_", ":", 1)[::-1] 
                            for x in self.registry.registry.keys() if x.startswith(guess)]
                logger.critical("Maybe you meant one of: {}".format(guesses))
                sys.exit(1)
        else:
            from damona.registry import Software
            r = Software(self.image_name).releases
            latest = r.last_release
            logger.info(f"No version found after {self.image_name} (e.g. fastqc:0.11.8)." + f" Installing latest version {latest}")
            registry_name = self.image_name
            # we look only at the prefix name, not the tag. so we should get the
            # registry names from self.registry that have the prefix in common, 
            # then extract the versions, and figure out the most recent.

        registry_name = self.registry.find_candidate(registry_name)
        if registry_name is None:
            logger.critical(f"No image found for {self.image_name}. Make sure it is correct using 'damona search' command")
            sys.exit(1)

        download_name = self.registry.registry[registry_name].download


        # get metadata from the registry recipe
        info = self.registry.registry[registry_name]
        logger.info(f"{info}")

        if info.binaries:
            self.binaries = info.binaries
        else:
            logger.warning(f"No binaries field found in registry of {registry_name}")
            self.binaries = [self.input_image.guessed_executable]

        # target file is stored in output_name variable
        if output_name is None:
            output_name = registry_name.replace(":", "_") + ".img"
        # here we check whether the image or binaries are already present.
        # if md5 is already provided, and image exists, nothing to copy
        if info.md5sum:
            if os.path.exists(self.images_directory / output_name):
                from easydev import md5
                md5_target = md5(self.images_directory / output_name)

                if md5_target == info.md5sum:
                    logger.info("Remote image and local image are identical, no need to download/pull again")
                    self.input_image = ImageReader(self.images_directory / output_name)
                    self.image_installed = True
                    return
        else:
            logger.warning(f"md5 field not found or not filled in {registry_name}." 
f"To be fixed in https://github.com/damona/damona/recipes/{self.image_name}/registry.yaml ")

        # now that we have the registry name, we can download the image
        logger.info("Downloading {}".format(download_name))

        # By default we downlaods from syslab.
        # if not found, one can provide an URL
        pull_folder = self.images_directory / "damona_buffer"

        if self.from_url:
            # The client does not support external https link other than
            # docker, library, shub.
            cmd = f"singularity pull --dir {pull_folder} "
            if force:
                cmd += " --force "
            cmd += f"{download_name}"
            subprocess.call(cmd.split())
        else:

            if download_name.startswith("https://"):
                print(f"downloading into {pull_folder} {output_name}")
                from urllib.request import urlretrieve
                urlretrieve(download_name, filename=str(pull_folder / output_name))
                # wget.download(download_name, str(pull_folder / output_name))

            else: # use singularity
                Client.pull(str(download_name), name=output_name,
                    pull_folder=pull_folder,
                    force=force)
        logger.info(f"File {self.image_name} uploaded to {pull_folder}")

        # Read the image, checking everything is correct

        self.input_image = ImageReader(pull_folder / output_name)
        shortname = self.input_image.shortname
        try:
            logger.info(f"Copying into damona image directory: {self.images_directory}")
            self.input_image.filename.rename(self.images_directory / shortname)
            self.input_image.filename = pathlib.Path(self.images_directory / shortname)
        except FileNotFoundError:
            logger.warning("File not installed properly. Stopping")
            self.image_installed = False

        # check the md5 validity
        if info.md5sum:
            if info.md5sum != self.input_image.md5:
                logger.warning(
"MD5 of downloaded image does not match the expected md5 found in the "
f"registry of {registry_name}. The latter may be incorrect in damona and needs "
f"to be updated in https://github.com/damona/damona/recipes/{self.image_name}/registry.yaml "
"or the donwload was interrupted")

        self.image_installed = True


    def install_binaries(self, force=False):
        if self.image_installed:
            logger.info("Creating binaries")
            bininst = BinaryInstaller(self.binaries, self.input_image.filename)
            bininst.install_binaries(force=force)
        else:
            logger.critical("The container image has not been installed. So, binaries cannot be installed either")


class BinaryInstaller():
    """Install a binary the bin/ directory of current environment given its image"""
    def __init__(self, binaries, parent_image_path):
        self.image = ImageReader(parent_image_path)
        self.binaries = binaries

    def install_binaries(self,  force=False):
        """Install an image and its binary

        """
        env = Environ()
        bin_directory = env.get_current_env() / "bin"

        for binary in sorted(self.binaries):
            bin_path = pathlib.Path(bin_directory) / binary

            cmd = """#!/bin/sh\nsingularity -s exec ${{DAMONA_SINGULARITY_OPTIONS}} {} {} ${{1+"$@"}}"""

            cmd = cmd.format(f"${{DAMONA_PATH}}/images/{self.image.shortname}", binary)

            if bin_path.exists() and force is False:
                name = pathlib.Path(bin_path).name
                path = str(bin_path).rstrip(bin_path.name)
                logger.warning(f"Binary {binary} exists already in {path}. Use --force to overwrite")
            else:
                # create binary and set correct permission
                with open(bin_path, "w") as fout:
                    fout.write(cmd)
                os.chmod(bin_path, 0o755)
                # aliases to print message
                name = pathlib.Path(bin_path).name
                path = str(bin_path).rstrip(bin_path.name)
                if bin_path.exists() and force is True:
                    logger.info(f"Re-installing '{binary}' in {path} since --force was used")
                else:
                    logger.info(f"Created binary {name} in {path}")

